
/**
 * typed-contracts v2.2.1
 * Validation with good Flow and TypeScript type inference
 */

const valueToString = value => {
  if (Number.isNaN(value)) return 'NaN';
  if (typeof value == 'function') return '[Function]'; // $FlowFixMe

  if (typeof value == 'symbol') return value.toString();

  switch (value) {
    case Infinity:
      return Infinity.toString();

    case -Infinity:
      return (-Infinity).toString();

    case undefined:
      return 'undefined';

    case null:
      return 'null';

    default:
      return `${JSON.stringify(value)} (${typeof value})`;
  }
};

const typesToString = expectedTypes => expectedTypes.slice(0, -2).concat(expectedTypes.slice(-2).join(' or ')).join(', ');

class ValidationError extends TypeError {
  /* :: +valueName: string; */

  /* :: +expectedTypes: $ReadOnlyArray<string>; */

  /* :: +value: mixed; */

  /* :: +nested: $ReadOnlyArray<ValidationError>; */
  constructor(valueName, value, expectedTypes = [], nested = []) {
    super();
    this.name = 'ValidationError';
    this.valueName = valueName;
    this.value = value;
    this.expectedTypes = [].concat(expectedTypes);
    this.nested = nested;

    if (this.expectedTypes.length) {
      this.message = this.nested.length ? `${this.expectedTypes[0]} ${this.valueName} have ${this.nested.length} nested validation errors` : `\`${this.valueName}\` must be ${typesToString(this.expectedTypes)}, but ${valueToString(value)} given`;
    }
  }

}

// eslint-disable-next-line no-redeclare
function curry2(fn) {
  return (...args) => args.length > 1 ? fn(args[0], args[1]) : value => fn(args[0], value);
}

// eslint-disable-next-line no-redeclare
function of(validate) {
  const maybe = curry2((valueName, value) => value == null ? value : validate(valueName, value));
  const optional = curry2((valueName, value) => value === undefined ? value : validate(valueName, value));

  const contract = (...args) => {
    if (args.length > 1) return validate(args[0], args[1]);

    const result = value => validate(args[0], value);

    result.maybe = maybe(args[0]);
    result.optional = optional(args[0]);
    return result;
  };

  const mapResult = transform => curry2((valueName, value) => transform(validate(valueName, value)));

  const match = (fromValue, fromError) => curry2((valueName, value) => {
    const result = validate(valueName, value);
    return result instanceof ValidationError ? fromError(result) : fromValue(result);
  });

  contract.maybe = maybe;
  contract.optional = optional;
  contract.mapResult = mapResult;
  contract.match = match;
  return contract;
}

const literal = expectedValue => of((valueName, value) => value === expectedValue || Number.isNaN(value) && Number.isNaN(expectedValue) ? value : new ValidationError(valueName, value, JSON.stringify(expectedValue)));
const isLiteral = literal;
const passLiteral = literal;
const lit = literal;
const isLit = literal;
const passLit = literal;

class UnionError extends ValidationError {
  constructor(valueName, value, errors) {
    super(valueName, value, 'Union', errors);
    this.name = 'UnionError';
  }

}
const union = (...rules) => {
  const validators = rules.map(rule => typeof rule == 'function' ? rule : literal(rule));

  const unionContract = (name, value) => {
    const errors = validators.map(validate => validate(name, value)).reduce((scope, error) => {
      if (error instanceof ValidationError) scope.push(error);
      return scope;
    }, []);
    return errors.length < validators.length ? value : new UnionError(name, value, errors);
  };

  return of(unionContract);
};
const isUnion = union;
const passUnion = union;
const uni = union;
const isUni = union;
const passUni = union;

class ArrayValidationError extends ValidationError {
  constructor(valueName, value, errors) {
    super(valueName, value, 'Array', errors);
    this.name = 'ArrayValidationError';
  }

}
const array = (...rules) => {
  const validate = rules.length === 1 ? rules[0] : union(...rules);
  return of((valueName, value) => {
    if (!Array.isArray(value)) {
      return new ValidationError(valueName, value, 'Array');
    }

    const errors = value.reduce((acc, item, index) => {
      const result = validate(`${valueName}[${index}]`, item);
      if (result instanceof ValidationError) acc.push(result);
      return acc;
    }, []);
    return errors.length ? new ArrayValidationError(valueName, value, errors) : value;
  });
};
const isArray = array;
const passArray = array;
const arr = array;
const isArr = array;
const passArr = array;

const boolean = of((name, value) => typeof value == 'boolean' ? value : new ValidationError(name, value, 'boolean'));
const isBoolean = boolean;
const passBoolean = boolean;
const bool = boolean;
const isBool = boolean;
const passBool = boolean;

const nul = of((name, value) => value === null ? value : new ValidationError(name, value, 'null'));
const isNull = nul;
const passNull = nul;

const number = of((name, value) => typeof value == 'number' ? value : new ValidationError(name, value, 'number'));
const isNumber = number;
const passNumber = number;
const num = number;
const isNum = number;
const passNum = number;

class ObjectValidationError extends ValidationError {
  constructor(valueName, value, errors) {
    super(valueName, value, 'Object', errors);
    this.name = 'ObjectValidationError';
  }

}

class MustBeExactError extends ValidationError {
  constructor(objectValueName, keyValueName, value) {
    super(`${objectValueName}.${keyValueName}`, value);
    this.name = 'MustBeExactError';
    this.message = `Object \`${objectValueName}\` must be exact but have excess \`${keyValueName}\` property`;
  }

}
function object(spec) {
  return of((valueName, value) => {
    if (!value || typeof value != 'object' || Array.isArray(value)) {
      return new ValidationError(valueName, value, 'Object');
    }

    const errors = Object.entries(spec).reduce((acc, [key, validate]) => {
      const result = validate(`${valueName}.${key}`, value[key]);
      if (result instanceof ValidationError) acc.push(result);
      return acc;
    }, []);
    return errors.length ? new ObjectValidationError(valueName, value, errors) : value;
  });
}
const isObject = object;
const passObject = object;
const obj = object;
const isObj = object;
const passObj = object;

function objectOf(...rules) {
  const validate = union(...rules);
  return of((valueName, value) => {
    if (!value || typeof value != 'object' || Array.isArray(value)) {
      return new ValidationError(valueName, value, 'Object');
    }

    const errors = Object.entries(value).reduce((acc, [key, val]) => {
      const result = validate(`${valueName}.${key}`, val);
      if (result instanceof ValidationError) acc.push(result);
      return acc;
    }, []);
    return errors.length ? new ObjectValidationError(valueName, value, errors) : value;
  });
}
const isObjectOf = objectOf;
const passObjectOf = objectOf;
const objOf = objectOf;
const isObjOf = objectOf;
const passObjOf = objectOf;

function shape(spec) {
  return object(Object.entries(spec).reduce((acc, [key, validate]) => {
    acc[key] = (valueName, value) => value === undefined ? value : validate(valueName, value);

    return acc;
  }, {}));
}

const string = of((name, value) => typeof value == 'string' ? value : new ValidationError(name, value, 'string'));
const isString = string;
const passString = string;
const str = string;
const isStr = string;
const passStr = string;

const undef = of((name, value) => value === undefined ? value : new ValidationError(name, value, 'void'));
const isUndefined = undef;
const passUndefined = undef;
const isUndef = undef;
const passUndef = undef;
const isVoid = undef;
const passVoid = undef;

export { ArrayValidationError, MustBeExactError, UnionError, ValidationError, arr, array, bool, boolean, isArr, isArray, isBool, isBoolean, isLit, isLiteral, isNull, isNum, isNumber, isObj, isObjOf, isObject, isObjectOf, isStr, isString, isUndef, isUndefined, isUni, isUnion, isVoid, lit, literal, nul, num, number, obj, objOf, object, objectOf, of, passArr, passArray, passBool, passBoolean, passLit, passLiteral, passNull, passNum, passNumber, passObj, passObjOf, passObject, passObjectOf, passStr, passString, passUndef, passUndefined, passUni, passUnion, passVoid, shape, str, string, undef, uni, union };
